/*
	PROforma, PROgs FOnt & Raster MAnager

	macros for defining multiplane bitmap drivers - separate planes
	- the number of planes has to be defined in PLANECOUNT
	- the depth of the plane should be defined in PLANEDEPTH
	- the single plane variant should be "BITMdriver init"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 5 November 1995
*/

extern BITMdriver init;

/* drawing in all the planes - using the pattern */

#define PLANE(local,colour,name,params,prot)								\
static Error colour##name##prot 											\
{                                                                           \
	int i; Error err;														\
	unsigned char *base=bitmap->base;										\
	unsigned char *pattern=bitmap->pattern; 								\
	bitmap->bitdraw=&init.##colour; 										\
	for (i=PLANECOUNT ; i ; 												\
		i--, bitmap->base+=bitmap->pinc, bitmap->pattern+=bitmap->pattpinc) \
		err=init.##colour.##name##params;									\
	bitmap->pattern=pattern;												\
	bitmap->base=base;														\
	bitmap->bitdraw=&##local.##colour;										\
	return err; 															\
}

/* drawing in the last plane only - no pattern */

#define PLANB(local,colour,name,params,prot)								\
static Error colour##name##prot 											\
{                                                                           \
	int i; Error err;														\
	unsigned char *base=bitmap->base;										\
	bitmap->bitdraw=&init.##colour; 										\
	bitmap->base+=bitmap->pinc*(PLANECOUNT-1);								\
	err=init.##colour.##name##params;										\
	bitmap->base=base;														\
	bitmap->bitdraw=&##local.##colour;										\
	return err; 															\
}

/* adjusting the pattern origin */

#define PLANO(local,colour,name)											\
static Error colour##name(Bitmap bitmap, short x, short y)					\
{                                                                           \
	int i;																	\
	unsigned char *pattern=bitmap->pattern; 								\
	int ppinc=bitmap->pattpinc; 											\
	short orgx=bitmap->patternbox.xorg; 									\
	short orgy=bitmap->patternbox.yorg; 									\
	for (i=PLANECOUNT ; i ; i--, bitmap->pattern+=ppinc)					\
	{                                                                       \
		bitmap->patternbox.xorg=orgx;										\
		bitmap->patternbox.yorg=orgy;										\
		init.##colour.ChangeOrigin(bitmap,x,y); 							\
	}																		\
	bitmap->pattern=pattern;												\
	return ERR_OK;															\
}

/* pixel read and write functions */

#define PIXREAD(first,others)												\
static Error first(Bitmap bitmap, int *colour)								\
{                                                                           \
	int i, pinc=bitmap->pinc;												\
	int newcol, totcol=0, shift=0; Error err;								\
	bitmap->cursorpos+=(PLANECOUNT-1)*pinc; 								\
	for (i=PLANECOUNT-1 ; i ;												\
		 i--, bitmap->cursorpos-=pinc, shift+=PLANEDEPTH)					\
	{                                                                       \
		init.##others(bitmap,&newcol);										\
		totcol=totcol+(newcol<<shift);										\
	}																		\
	err=init.first(bitmap,&newcol); 										\
	*colour=totcol+(newcol<<shift); 										\
	return err; 															\
}

#define PIXSET(first,others)												\
static Error first(Bitmap bitmap, int colour)								\
{                                                                           \
	int i, pinc=bitmap->pinc;												\
	bitmap->cursorpos+=(PLANECOUNT-1)*pinc; 								\
	for (i=PLANECOUNT-1 ; i ; i--, bitmap->cursorpos-=pinc) 				\
	{                                                                       \
		init.##others(bitmap,colour);										\
		colour>>=PLANEDEPTH;												\
	}																		\
	return init.first(bitmap,colour);										\
}
