/*
	PROforma, PROGS Font & Raster Manager

	PFPGstateReAssign
		reassign a gstate to a different driver
		only PF_DRIVER_SCREEN and PF_DRIVER_DUMMY are accepted drivers !!
		The gstate is released if an error occurs !!

	The area which is opened for a driver is clipped to the imageable area.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 21 March 1996
	06.08.1996 now the ColourCache is also allocated and initialised
	19.09.1996 ColourCacheInit crashed on <68030 because pattsize not even
	12.02.1998 PFPGstateReAssign2 added (need possibility to set the device)
*/

#include "gstate.h"

extern DRIVdriver *screen;		/* current screen driver */
extern DRIVdriver dummydriver;
extern BITMdriver *bitmaps; 	/* currently supported bitmap drivers */
extern font *ResFonts;			/* builtin font */
extern int ColourCacheCount;

Error ColourCacheInit(Gstate gstate, ColourCache *base, Size pattsize, int count);

Error PFPGstateReAssign(Gstate gstate, int driverID,
						short xsiz, short ysiz, short xorg, short yorg)
{
	return PFPGstateReAssign2(gstate, driverID, NULL, xsiz, ysiz, xorg, yorg);
}

Error PFPGstateReAssign2(Gstate gstate, int driverID, char *device,
						short xsiz, short ysiz, short xorg, short yorg)
{
	Error err;
	DRIVdriver *driver=&dummydriver;
	BITMdriver *bdriver;
	PBbox clip;

	/* the given driverID has to be retrieved in the driverlist */

	if (driverID==PF_DRIVER_SCREEN)
		driver=screen;

	/* find the bitmap driver which the printer driver needs */
	for (bdriver=bitmaps ;
		 bdriver && !MEMSame(sizeof(BitmapType),
							 &bdriver->type,&driver->bitmap) ;
		 bdriver=bdriver->next);
	if (bdriver==NULL) { PFGstateRemove(gstate); return ERR_IEXP; }

	/* release old printer driver stuff */
	gstate->printer->CleanUp(gstate);
	if (gstate->BitmapAllocated)
	{
		MEMRelease(gstate->rasterbase);
		gstate->BitmapAllocated=FALSE;
	}
	MEMRelease(gstate->rasterpattern);

	gstate->printer=driver; 		/* set printer driver for gstate */
	gstate->bitmap=bdriver; 		/* set bitmap driver for gstate */

	/* start initialising Gstate */
	/* the driver has to - open device for display */
	/*					 - determine conversion factors todev and ppp */
	/*					 - fill in maximum area which can be used (clip) */
	err=driver->InitInit(gstate, device, &clip);
	if (err)
	{
		PFGstateRemove(gstate); 	/* release Gstate again */
		return err; 				/* report error */
	}

	/* fill in orgppp and orgtodev */
	MEMCopy((long)sizeof(coor),&gstate->ppp,&gstate->orgppp);
	MEMCopy((long)sizeof(coor),&gstate->todev,&gstate->orgtodev);

	/* clip the requested area to the useable part */
	xorg-=clip.xorg;
	yorg-=clip.yorg;
	if (xorg<0) { xsiz+=xorg; xorg=0; }
	if (yorg<0) { ysiz+=yorg; yorg=0; }
	if (xsiz>clip.xsiz-xorg) xsiz=clip.xsiz-xorg;
	if (ysiz>clip.ysiz-yorg) ysiz=clip.ysiz-yorg;
	if (driver!=&dummydriver && (xsiz<=0 || ysiz<=0))	/* nothing to draw */
		{ PFGstateRemove(gstate); return ERR_ORNG; }

	/* set pagebbox of entire page (is cleared before) */
	gstate->allbbox.xsiz = xsiz;
	gstate->allbbox.ysiz = ysiz;
	gstate->allbbox.xporg= short2pt(xorg+clip.xorg);
	gstate->allbbox.yporg= short2pt(yorg+clip.yorg);

	/* we will now try to allocate the bitmap to render the page in */
	{
		Size buffsize;
		Size maxsize;				/* approximation of free memory */
		char *base;
		Bitmap raster;

		driver->BufferSize(gstate,xsiz,ysiz,&buffsize);
		gstate->passes=1;				/* start with one run */
		/* now allocate buffer, FastPattern (& clear) */
		raster=gstate->raster;
		if (buffsize)
		{
			Size linc;
			err=MEMAllocate(buffsize,&raster->base);
			if (err) { PFGstateRemove(gstate); return ERR_IMEM; }
			gstate->rasterbase=raster->base;
			gstate->BitmapAllocated=TRUE;
			raster->clip.xsiz=raster->all.xsiz=xsiz;
			raster->clip.ysiz=raster->all.ysiz=ysiz;
			bdriver->BufferSize(xsiz,1,&linc);
			if (bdriver->type.PlaneGrouping==BITMAP_SEPARATE_PLANES)
			{
				linc/=bdriver->type.PlaneCount;
				raster->linc=linc;
				raster->pinc=linc*ysiz;
			}
			else
				raster->linc=linc;
		}
		{
			Size pattsize, linc;
			ColourCache *base;
			int count=ColourCacheCount;
			
			bdriver->BufferSize(bdriver->FastPatternXsize,1,&linc);
			if (bdriver->type.PlaneGrouping==BITMAP_SEPARATE_PLANES)
			{
				linc/=bdriver->type.PlaneCount;
				raster->pattlinc=linc;
				raster->pattpinc=linc*bdriver->FastPatternYsize;
			}
			else
				raster->pattlinc=linc;
			bdriver->BufferSize(bdriver->FastPatternXsize,
								bdriver->FastPatternYsize,&pattsize);
			pattsize=even(pattsize);
			err=MEMAllocate(pattsize*count+sizeof(ColourCache)*(count+1),&base);
			if (err) { PFGstateRemove(gstate); return ERR_IMEM; }
			raster->pattern=gstate->rasterpattern=(char *)base;
			raster->patternbox.xsiz=bdriver->FastPatternXsize;
			raster->patternbox.ysiz=bdriver->FastPatternYsize;
			/* initialise ColourCache */
			ColourCacheInit(gstate,base,pattsize,count);
		}

		/* set PageBbox & firstbbox */
		MEMCopy(sizeof(PPageBbox),&gstate->allbbox,&gstate->pagebbox);
		gstate->pagebbox.ysiz=ysiz;
		MEMCopy(sizeof(PPageBbox),&gstate->pagebbox,&gstate->firstbbox);
	}

	/* finalise initialisation, this may */
	/*		- modify the pagebbox, allbbox and firstbbox */
	/*		- modify the use of the raster buffer for eg multiple buffers,
	/*		  add header, use part to render a line etc. */
	err=driver->InitDone(gstate);
	if (err)
	{
		PFGstateRemove(gstate); 	/* release Gstate again */
		return err; 				/* report error */
	}

	PFCTMReset(gstate); 			/* reset CTM */
	PFPageBboxReset(gstate);		/* reset pagebbox */
	PFColourGray(gstate,pt_hundred);/* set the drawing colour */

	return ERR_OK;
}
