/*
	PROforma, PROgs FOnt & Raster MAnager

	Routines to set some general parameters in the Gstate

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 July 1993
	09.11.1995 linewidth is now forced in range instead of producing an error
	21.06.1996 PFDisplayMode added
	24.06.1996 PFLineWidth now sets the linewidth which will be used. The line
			   width is no longer affected by changes to CTM of PageScale. Line
			   end is always round, the circle can be scaled, but not rotated.
	24.11.1996 - calling PFPathStroked after PFLineWidth always reset to
				 hairline drawing - fixed (hairline was not reset to FALSE when needed)
			   - PFLineWidth caused problems when scale factor<1
*/

#include <math.h>

#include "gstate.h"
#include "system.h"

/* routines to control the handling of text */
Error PFPrintEscape(Gstate gstate, int status)
{
	gstate->printescape=status;
	return ERR_OK;
}
Error PFExtraEOS(Gstate gstate, int character)
{
	gstate->extraeos=character;
	return ERR_OK;
}

/* routines to set the path type */

Error PFPathMethod(Gstate gstate, int method)
{
	gstate->pathtype= method==PF_PATH_STROKED ?
				( gstate->hairline ? PF_PATH_HAIRLINE : PF_PATH_THICKLINE ) :
				method;
	return ERR_OK;
}

/* routine to set the linewidth */

Error PFLineWidth(Gstate gstate, pt width)
{
#define square(a) fixmul((a),(a))
	pt xscale, yscale;
	pt linexppp, lineyppp;
	pt linexwid, lineywid;

	if (width>short2pt(500)) width=short2pt(500);
	else if (width<0) width=0;
	gstate->linewidth=width/2;

	/* speedy handling of zero linewidth -> force hairline */
	if (width==0)
	{
		gstate->hairline=TRUE;
		if (gstate->pathtype==PF_PATH_THICKLINE)
			gstate->pathtype=PF_PATH_HAIRLINE;
		return ERR_OK;
	}

	/* calculate linexppp and lineyppp */
	/* these are scaled versions of ppp (pixels to point) */
	/* which are scaled by the scale factors from the CTM and also */
	/* by the current linewidth. This means that the linewidth is always 1 */
	if (gstate->todefaultCTM.xr==0 && gstate->todefaultCTM.yr==0)
	{
		xscale=gstate->todefaultCTM.xs;
		yscale=gstate->todefaultCTM.ys;
	}
	else
	{
		/* BUGGY - THIS IS COMPLETELY WRONG !!! */
		xscale=fixsqrt( square(gstate->todefaultCTM.xs/64) +
						square(gstate->todefaultCTM.yr/64) )*64;
		yscale=fixsqrt( square(gstate->todefaultCTM.xr/64) +
						square(gstate->todefaultCTM.ys/64) )*64;
	}
	if (xscale<0) xscale=-xscale;
	if (yscale<0) yscale=-yscale;
	gstate->linexppp=fixmul(xscale,gstate->ppp.x);
	gstate->lineyppp=fixmul(yscale,gstate->ppp.y);
	if (gstate->linexppp==gstate->lineyppp)
		gstate->uniformgrid=TRUE;
	else
	{
		gstate->uniformgrid=FALSE;
		gstate->liney2x=fixdiv(gstate->lineyppp,gstate->linexppp);
		gstate->linex2y=fixdiv(gstate->linexppp,gstate->lineyppp);
	}
	linexwid=fixmul(width,fixmul(xscale,gstate->todev.x));
	lineywid=fixmul(width,fixmul(yscale,gstate->todev.y));
	if (linexwid<pt_one && lineywid<pt_one)
	{
		gstate->hairline=TRUE;
		if (gstate->pathtype==PF_PATH_THICKLINE)
			gstate->pathtype=PF_PATH_HAIRLINE;
		return ERR_OK;
	}
	gstate->linexmid= (pt2rint(linexwid)&1) ? TRUE : FALSE;
	gstate->lineymid= (pt2rint(lineywid)&1) ? TRUE : FALSE;
	gstate->hairline=FALSE;
	if (gstate->pathtype==PF_PATH_HAIRLINE)
			gstate->pathtype=PF_PATH_THICKLINE;
	return ERR_OK;
}

/* routine to set the flatness */

Error PFFlatness(Gstate gstate, pt flat)
{
	if (flat<(pt_one/5)) flat=pt_one/5;
	if (flat>pt_hundred) flat=pt_hundred;
	gstate->flatness=flat;
	return ERR_OK;
}

/* routine to scale the page !? */
Error PFPageScale(Gstate gstate, pt factor)
{
	pt inverse;
	factor=gstate->Scale=fixmul(gstate->Scale,factor);
	inverse=fixdiv(pt_one,factor);
	gstate->todev.x=fixmul(gstate->orgtodev.x,factor);
	gstate->todev.y=fixmul(gstate->orgtodev.y,factor);
	gstate->ppp.x=fixmul(gstate->orgppp.x,inverse);
	gstate->ppp.y=fixmul(gstate->orgppp.y,inverse);

	/* currCTM should also convert to device coordinates !! */
	return currCTMset(gstate);
}

Error PFDisplayMode(Gstate gstate, int mode)
{
	return gstate->printer->SetDisplayMode(gstate,mode);
}
