/*
	PROforma, PROgs FOnt & Raster MAnager

	Fill routine

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 6 August 1993
	19.03.1994 adjusted for (external) drivers
	29.11.1995 attempt to make calc_nexty more efficient
	30.11.1996 nexthxpos no longer has to adjust half a pixel
	01.12.1996 release remains of ylist, could cause memory leaks
*/

#include "gstate.h"
#include "system.h"
#include "path.h"

/* forward declarations */
static segment *add_xlist(segment *ylist, segment *xlist, short *nexty);
static void calc_nexty(segment *s, short *nexty);

#define nexthxpos(a) (a->startx/*+(a->incr/2)*/)
#define nextxpos(a) (a->startx+a->incr)

/* ACTUAL DRAWING COMMAND */

/*
	Filling algorithme

	needs a list in which all line segments are order by increasing starty
	during the execution of this command active lines are moved from this list
	(ylist) to the xlist, in which lines are stored from left to right.
	This list is used to determine the trapezoids which build the filled and
	clipped path.
	The main complication in this algorithme are intersecting lines. This is
	solved by actually processing bands with as event points (y coordinates)
	start, end and intersection points. If there was an intersection then the
	xlist has to be resorted.
	The error code is always ERR_OK, except when nothin has been drawn, then
	its ERR_ORNG (for fast picture drawing).
*/

Error fillpath(Gstate gstate,
			  Error (*Trapezoid)(Bitmap, short, short, pt, pt, pt, pt),
			  Bitmap bitmap, int rule)
{
	segment *sh, *ss;
	segment *xlist, *ylist; 	/* xlist, last element in xlist, ylist */
	segment *el, *next; 		/* used for running through lists */
	short boty; 				/* top and bottom value of PageBbox */
	short starty, endy, nexty;	/* band boundaries and next event point */
	short bandheight;			/* = endy-starty */
	short wind, clipwind;		/* counter to determine in or outside */
	short leftside; 			/* TRUE if leftside is initialised */
	pt leftstartx, leftincr;	/* parameters for leftside of trapezoid */
	Error err=ERR_ORNG; 		/* assume nothing to draw */

	/* we have to remove all the segments which fall above or under the
	   PageBbox (y values too high or too low) */
	el=gstate->path->next;
	boty=gstate->raster->clip.ysiz;

	while (el->next)
	{
		next=el->next;
		if (el->starty>=boty || el->endy<0)
		{
			/* unlink and free item from path */
			el->next->prev=el->prev;
			el->prev->next=el->next;
			free_segment(gstate,el);
		}
		el=next;
	}

	/* remove the header and sentinel segments, and use them to initialise the
	   xlist, we only need the linkage with next in the ylist */
	xlist=sh=gstate->path;
	ss=sh->prev;
	ylist=sh->next;
	if (!ylist) return err; 		/* there has to be something to do */
	sh->next=ss; sh->prev=NULL; 	/* has to be zero for calc_nexty to work */
	ss->prev->next=NULL;
	ss->prev=sh;

	/* start building xlist by filling in all segment with same starty */
	nexty=boty; 					/* maximum for next value */
	endy=ylist->starty; 			/* start of first band */
	ylist=add_xlist(ylist, xlist, &nexty);

#ifdef DEBUG
puts("process bands inside PageBbox");
#endif

	/* process all the bands inside PageBbox */
	while(endy<boty)
	{

#ifdef DEBUG			/* write out the xlist */
el=xlist->next; 		/* skip header */
while (el->next)
{
	printf("xlist startx %x incr %x starty %d nexty %d endy %d clip %d        ",
	el->startx, el->incr, el->starty, el->nexty, el->endy, el->clip);
	el=el->next;
}
#endif

		if (ylist) nexty=min(nexty,ylist->starty);
										/* determine first event point */

		starty=endy;					/* initialise this band */
		endy=nexty;
		bandheight=endy-starty;
		nexty=boty;
		wind=0; clipwind=0; leftside=FALSE;

#ifdef DEBUG
printf("band starty %d, endy %d\n", starty, endy);
#endif

		/* now run through the xlist */
		el=xlist->next; 				/* skip header */
		while (el->next)
		{
			next=el->next;				/* keep next in case of removal */

			/* adjust counter for window rule */
			if (el->isclip) clipwind+=el->direction;
					   else wind+=el->direction;

#ifdef DEBUG
printf("winding counter %d, clip %d ",wind,clipwind);
#endif

			/* is this a leftside ?? */
			if (!leftside && wind&rule && clipwind&rule)
			{
				leftside=TRUE;
				leftstartx=el->startx;
				leftincr=el->incr;
#ifdef DEBUG
printf("leftside start %x, increment %x ",leftstartx,leftincr);
#endif
			}

			/* is this a rightside ?? */
			if (leftside && !(wind&rule && clipwind&rule) )
			{
#ifdef DEBUG
printf("rightside start %x, increment %x ",el->startx,el->incr);
#endif
				err=Trapezoid(bitmap, starty, endy, leftstartx, leftincr,
							   el->startx, el->incr);
				leftside=FALSE;
			}

#ifdef DEBUG
puts("; reduce band from element");
#endif

			/* now reduce band from this element */
			el->starty=endy;
			el->startx+=el->incr*bandheight;
			if (endy>=el->endy)
			{
				/* unlink and free element from xlist */
				/* also update nexty for the element which has a new prev */
				el->next->prev=el->prev;
				el->prev->next=el->next;
				free_segment(gstate,el);
#ifdef DEBUG
printf(" element removed ");
#endif
			} else {
				/* recalculate and update nexty, can be surpassed, or changed */
				if (el->nexty<=endy) el->nexty=el->endy;
				calc_nexty(el, &nexty);

				/* re-order this element in xlist */
				/* this involves recalculating all elements with a new prev */

				if ( nextxpos(el->prev) > nextxpos(el) )
				{
					segment *s=el, *list=el->next;
					pt x=nexthxpos(s);

					/* linkout first */
					s->next->prev=s->prev;
					s->prev->next=s->next;
					el=s->prev; 	/* make sure continue with correct segment */

					/* relink in place now */
					while (nexthxpos(list->prev)>x) list=list->prev;
					/* link in before segment *list */
					s->prev=list->prev;
					s->next=list;
					s->prev->next=s;
					list->prev=s;
					calc_nexty(list, &nexty);
					calc_nexty(s, &nexty);
				}
#ifdef DEBUG
printf("nexty=%d ",nexty);
#endif
			}
			el=next;
		}

/* just a test, render each line individually
nexty=endy+1; */
		if (!ylist && !xlist->next->next) nexty=boty+1; /* x,ylist empty->stop */

		if (ylist && endy==ylist->starty) ylist=add_xlist(ylist, xlist, &nexty);
	}

	/* release some remaining path */
	for ( ; ylist ; ylist=next)
	{
		next=ylist->next;
		free_segment(gstate,ylist);
	}

	/* now clean up some internal data */
	gstate->path=xlist; 	  /* set path to be released, not valid !! */
	return err;
}

static segment *add_xlist(segment *ylist, segment *xlist, short *nexty)
{
	short y;
	segment *s;
	pt x;

	if (ylist) y=ylist->starty;
	while(ylist && ylist->starty==y)
	{
		linkout(s,ylist);
		/* now link s in xlist */
		x=nexthxpos(s);
		/* find item before which we have to link in */
		while (nexthxpos(xlist)>x) xlist=xlist->prev;
		while (nexthxpos(xlist)<x) xlist=xlist->next;
		/* link in before segment *xlist */
		s->prev=xlist->prev;
		s->next=xlist;
		s->prev->next=s;
		xlist->prev=s;
		/* determine nexty */
		s->nexty=s->endy;
		calc_nexty(s, nexty);
		calc_nexty(s->next, nexty);
	}
	return ylist;
}

/* calculate intersection of an element with it's predecessor */
/* nexty should be adjusted after the change */
static void calc_nexty(segment *s, short *nexty)
{
	short addy; 			/* offset from current y position to crossing */
	segment *p=s->prev; 	/* previous element (to test intersection with) */
#ifdef OLDCODE
	pt divisor=s->incr - p->incr;

	/* don't do nothing if there is no preceding segment */
	if (p->prev && divisor!=0)
	{
		addy= ( p->startx - s->startx + (divisor/2) ) / ( divisor );
		if (addy>0) s->nexty=min(s->nexty, s->starty+addy);
	}
	*nexty=min(*nexty,s->nexty);
#else
	unsigned long divisor; /* actually a pt */

	/* don't do nothing if there is no preceding segment */
	if (p->prev && p->incr>s->incr)
	{
		divisor=p->incr-s->incr;
		addy= (unsigned)( s->startx - p->startx + (divisor/2) ) / ( divisor );
		if (addy>0) s->nexty=min(s->nexty, s->starty+addy);
	}
	*nexty=min(*nexty,s->nexty);
#endif OLDCODE
#ifdef DEBUG
printf(" | addy=%d, endy=%d, nexty=%d | ",addy,s->endy,s->nexty);
#endif
}
