/*
	PROforma, PROgs FOnt & Raster MAnager

	PFGstateSave
	PFGstateRestore
		Save and restore all the graphical settings in the gstate. This allows
		you to create drawing entities which do not change anything in the
		drawing environment.
		Entitities which are saved include :
			CTM, linewidth, paper and drawing colour, clipping path, current
			font (when it remains loaded), fontsize, pagebbox and origin,
			flatness and current position, path drawing method, usecache,
			extraeos, printescape,...

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	10.11.1996 started
	27.11.1996 bug with nesting of fixed
*/

#include "gstate.h"
#include "system.h"
#include "list.h"

Error PFGstateSave(Gstate gstate)
{
	gstatelist *l;
	Error err;

	if (gstate->freesavelist)
		linkout(l,gstate->freesavelist);
	else
		catch( MEMAllocate(sizeof(gstatelist),&l) );

	gstate->rasterclip=gstate->raster->clip;
	MEMCopy(sizeof(gstatelist),&gstate->next,l);
	gstate->clippathmine=FALSE;

	linkin(l,gstate->savelist);
	return ERR_OK;
}

Error PFGstateRestore(Gstate gstate)
{
	gstatelist *l;
	pt fs;

	if (!gstate->savelist) return ERR_IPAR;
	linkout(l,gstate->savelist);

	/* release memory used by clipping path */
	if (gstate->clippathmine) ReleaseClipPath(gstate);

	/* swap areas - also unlinks in the process */
	MEMSwap(sizeof(gstatelist),l,&gstate->next);

	/* now make sure the other parts of the gstate are also ok again */

	gstate->raster->clip=gstate->rasterclip;

	if (!MEMSame(sizeof(CTM),&l->todefaultCTM,&gstate->todefaultCTM))
		currCTMset(gstate);

	fs=gstate->fontsize;
	gstate->fontsize=l->fontsize;
	PFFontScale(gstate,fs);
	PFFontSelect(gstate,gstate->currfontname);

	switch (gstate->drawcolourspace)
	{
		case PF_COLOURSPACE_GRAYSHADE:
			PFColourGray(gstate,gstate->DrawColour.black);
			break;
		case PF_COLOURSPACE_RGB:
			PFColourRGB(gstate,(ColourRGB *)&gstate->DrawColour);
			break;
		case PF_COLOURSPACE_CMYK:
			PFColourCMYK(gstate,&gstate->DrawColour);
			break;
	}

	linkin(l,gstate->freesavelist);
	return ERR_OK;
}
